/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import org.ow2.orchestra.bpmn2bpel.exceptions.TransformerException;
import org.ow2.orchestra.jaxb.bpel.TActivity;


/**
 * @author Guillaume Porcher
 */
public final class BpelUtil {

  private BpelUtil() {
  }

  /**
   * Set child activity according to its type
   * (calls setEmpty, setInvoke, setReceive...)
   *
   * @param parentActivity
   * @param childActivity
   */
  public static void setChildActivity(final Object parentActivity, final TActivity childActivity) {
    if (parentActivity == null) {
      throw new TransformerException("Error while setting child activity: parent activity is null");
    }
    if (childActivity == null) {
      throw new TransformerException("Error while setting child activity: child activity is null");
    }
    final String methodName = "set" + childActivity.getClass().getSimpleName();
    try {
      final Method m = parentActivity.getClass().getMethod(methodName, childActivity.getClass());
      m.invoke(parentActivity, childActivity);
    }
    catch (final InvocationTargetException e) {
      throw new TransformerException("error while setting activity", e);
    }
    catch (final SecurityException e) {
      throw new TransformerException("error while setting activity", e);
    }
    catch (final NoSuchMethodException e) {
      throw new TransformerException("error while setting activity", e);
    }
    catch (final IllegalArgumentException e) {
      throw new TransformerException("error while setting activity", e);
    }
    catch (final IllegalAccessException e) {
      throw new TransformerException("error while setting activity", e);
    }
  }

}
