/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import javax.xml.transform.dom.DOMResult;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;

import org.ow2.orchestra.jaxb.b4p.PeopleActivity;
import org.ow2.orchestra.jaxb.bpel.Process;
import org.ow2.orchestra.jaxb.bpmn.Definitions;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;


/**
 * @author Guillaume Porcher
 */
public final class JaxbUtil {

  private JaxbUtil() {
  }

  /**
   * Parse BPMN document from input stream
   *
   * @param inputStream the BPMN XML document
   * @return the BPMN definitions object
   * @throws JAXBException
   * @throws SAXException
   */
  public static Definitions parseBPMNDefinitions(final InputStream inputStream) throws JAXBException, SAXException {
    final JAXBContext jc = JAXBContext.newInstance("org.ow2.orchestra.jaxb.bpmn");
    final Unmarshaller u = jc.createUnmarshaller();

    final SchemaFactory sf = SchemaFactory.newInstance(javax.xml.XMLConstants.W3C_XML_SCHEMA_NS_URI);
    final Schema schema = sf.newSchema(JaxbUtil.class.getClassLoader().getResource("BPMN20.xsd"));
    u.setSchema(schema);
    return (Definitions) u.unmarshal(inputStream);
  }

  /**
   * Generate XML document from java Bpel process
   *
   * @param bpelProcess  the bpel process object
   * @param outputStream outputstream where the BPEL XML will be written.
   * @throws JAXBException
   * @throws IOException
   * @throws SAXException
   */
  public static void writeBPEL(
          final Process bpelProcess,
          final OutputStream outputStream)
          throws JAXBException, IOException, SAXException {
    final JAXBContext context = JAXBContext.newInstance("org.ow2.orchestra.jaxb.bpel");
    final Marshaller m = context.createMarshaller();
    final SchemaFactory sf = SchemaFactory.newInstance(javax.xml.XMLConstants.W3C_XML_SCHEMA_NS_URI);
    final Schema schema = sf.newSchema(JaxbUtil.class.getClassLoader().getResource("ws-bpel_executable.xsd"));
    m.setSchema(schema);
    DOMResult domResult = new DOMResult();
    m.marshal(bpelProcess, domResult);
    Node bpelXmlNode = domResult.getNode();
    String bpelXml = XmlUtil.toString(bpelXmlNode);
    outputStream.write(bpelXml.getBytes());
    outputStream.flush();
  }

  public static Element writePeopleActivity(
          final PeopleActivity peopleActivity)
          throws JAXBException, IOException, SAXException {
    final JAXBContext context = JAXBContext.newInstance("org.ow2.orchestra.jaxb.b4p");
    final Marshaller m = context.createMarshaller();
    final SchemaFactory sf = SchemaFactory.newInstance(javax.xml.XMLConstants.W3C_XML_SCHEMA_NS_URI);
    final Schema schema = sf.newSchema(JaxbUtil.class.getClassLoader().getResource("ws-bpel4people.xsd"));
    m.setSchema(schema);
    DOMResult domResult = new DOMResult();
    m.marshal(peopleActivity, domResult);
    Node xmlNode = domResult.getNode();
    return ((Document) xmlNode).getDocumentElement();
  }
}
