/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel.components;

import java.util.ArrayList;
import java.util.List;
import org.ow2.orchestra.bpmn2bpel.BpmnMap;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;


/**
 * Represent a sequence component
 *
 * @author Guillaume Porcher
 */
public class SequenceComponent extends AbstractComponent {

  private final List<TFlowNode> sequenceNodes;

  public SequenceComponent(final List<TFlowNode> nodes, final BpmnMap bpmnMap) {
    this.sequenceNodes = new ArrayList<TFlowNode>();
    this.fillSequenceNodes(nodes, bpmnMap);
  }

  /**
   * Add nodes to sequenceNodes.
   * If a node to add is a sequenceComponent, don't add the component but add nested nodes.
   * (merge nested sequences)
   *
   * @param nodes
   */
  private void fillSequenceNodes(final List<TFlowNode> nodes, final BpmnMap bpmnMap) {
    for (final TFlowNode node : nodes) {
      if ((node instanceof SequenceComponent)
              && bpmnMap.getIncomingLinks(node).isEmpty()
              && bpmnMap.getOutgoingLinks(node).isEmpty()) {
        this.fillSequenceNodes(((SequenceComponent) node).getSequenceNodes(), bpmnMap);
      } else {
        this.sequenceNodes.add(node);
      }
    }
  }

  /**
   * @return the sequenceNodes
   */
  public List<TFlowNode> getSequenceNodes() {
    return this.sequenceNodes;
  }

  @Override
  public String toString() {
    final StringBuilder sb = new StringBuilder();
    sb.append(this.getClass().getSimpleName());
    sb.append("[").append(this.getId()).append("]");
    sb.append(" (");
    for (final TFlowNode flowNode : this.sequenceNodes) {
      sb.append(flowNode.getId()).append(", ");
    }
    sb.delete(sb.length() - 2, sb.length());
    sb.append(")");
    return sb.toString();
  }
}
