/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel.generators;


import org.ow2.orchestra.bpmn2bpel.BpmnMap;
import org.ow2.orchestra.bpmn2bpel.WSDLGenerator;
import org.ow2.orchestra.bpmn2bpel.components.EmptyComponent;
import org.ow2.orchestra.bpmn2bpel.components.FlowComponent;
import org.ow2.orchestra.bpmn2bpel.components.JoinConditionComponent;
import org.ow2.orchestra.bpmn2bpel.components.LinkComponent;
import org.ow2.orchestra.bpmn2bpel.components.OnEventComponent;
import org.ow2.orchestra.bpmn2bpel.components.PickComponent;
import org.ow2.orchestra.bpmn2bpel.components.RepeatComponent;
import org.ow2.orchestra.bpmn2bpel.components.RepeatWhileComponent;
import org.ow2.orchestra.bpmn2bpel.components.SequenceComponent;
import org.ow2.orchestra.bpmn2bpel.components.SwitchComponent;
import org.ow2.orchestra.bpmn2bpel.components.WhileComponent;
import org.ow2.orchestra.bpmn2bpel.exceptions.TransformerException;
import org.ow2.orchestra.jaxb.bpel.Assign;
import org.ow2.orchestra.jaxb.bpel.Copy;
import org.ow2.orchestra.jaxb.bpel.Documentation;
import org.ow2.orchestra.jaxb.bpel.Empty;
import org.ow2.orchestra.jaxb.bpel.Extension;
import org.ow2.orchestra.jaxb.bpel.Extensions;
import org.ow2.orchestra.jaxb.bpel.Flow;
import org.ow2.orchestra.jaxb.bpel.From;
import org.ow2.orchestra.jaxb.bpel.FromPart;
import org.ow2.orchestra.jaxb.bpel.FromParts;
import org.ow2.orchestra.jaxb.bpel.Import;
import org.ow2.orchestra.jaxb.bpel.Link;
import org.ow2.orchestra.jaxb.bpel.Links;
import org.ow2.orchestra.jaxb.bpel.PartnerLink;
import org.ow2.orchestra.jaxb.bpel.PartnerLinks;
import org.ow2.orchestra.jaxb.bpel.Process;
import org.ow2.orchestra.jaxb.bpel.Sequence;
import org.ow2.orchestra.jaxb.bpel.Source;
import org.ow2.orchestra.jaxb.bpel.Sources;
import org.ow2.orchestra.jaxb.bpel.TActivity;
import org.ow2.orchestra.jaxb.bpel.TBoolean;
import org.ow2.orchestra.jaxb.bpel.TCondition;
import org.ow2.orchestra.jaxb.bpel.Target;
import org.ow2.orchestra.jaxb.bpel.Targets;
import org.ow2.orchestra.jaxb.bpel.To;
import org.ow2.orchestra.jaxb.bpel.ToPart;
import org.ow2.orchestra.jaxb.bpel.ToParts;
import org.ow2.orchestra.jaxb.bpel.Variable;
import org.ow2.orchestra.jaxb.bpel.Variables;
import org.ow2.orchestra.jaxb.bpmn.Assignment;
import org.ow2.orchestra.jaxb.bpmn.DataInput;
import org.ow2.orchestra.jaxb.bpmn.DataInputAssociation;
import org.ow2.orchestra.jaxb.bpmn.DataOutput;
import org.ow2.orchestra.jaxb.bpmn.DataOutputAssociation;
import org.ow2.orchestra.jaxb.bpmn.Operation;
import org.ow2.orchestra.jaxb.bpmn.Participant;
import org.ow2.orchestra.jaxb.bpmn.Property;
import org.ow2.orchestra.jaxb.bpmn.TCatchEvent;
import org.ow2.orchestra.jaxb.bpmn.TDataObject;
import org.ow2.orchestra.jaxb.bpmn.TEventDefinition;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;
import org.ow2.orchestra.jaxb.bpmn.TInterface;
import org.ow2.orchestra.jaxb.bpmn.TItemDefinition;
import org.ow2.orchestra.jaxb.bpmn.TMessage;
import org.ow2.orchestra.jaxb.bpmn.TMessageEventDefinition;
import org.ow2.orchestra.jaxb.bpmn.TProcess;
import org.ow2.orchestra.jaxb.bpmn.TServiceTask;
import org.ow2.orchestra.jaxb.bpmn.TThrowEvent;
import org.ow2.orchestra.jaxb.bpmn.TTimerEventDefinition;
import org.ow2.orchestra.jaxb.bpmn.TUserTask;
import org.ow2.orchestra.util.Misc;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

import javax.wsdl.Definition;
import javax.xml.XMLConstants;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.namespace.QName;
import java.io.IOException;
import java.util.Collection;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Map.Entry;


/**
 * @author Guillaume Porcher
 */
public final class BpelGenerator {

  private final PickGenerator pickGenerator = new PickGenerator(this);
  private final FlowGenerator flowGenerator = new FlowGenerator(this);
  private final IfGenerator ifGenerator = new IfGenerator(this);
  private final InvokeGenerator invokeGenerator = new InvokeGenerator(this);
  private final ReceiveGenerator receiveGenerator = new ReceiveGenerator(this);
  private final ReplyGenerator replyGenerator = new ReplyGenerator(this);
  private final SequenceGenerator sequenceGenerator = new SequenceGenerator(this);
  private final WaitGenerator waitGenerator = new WaitGenerator(this);
  private final WhileGenerator whileGenerator = new WhileGenerator(this);
  private final OnEventGenerator onEventGenerator = new OnEventGenerator(this);
  private final PeopleActivityGenerator peopleActivityGenerator = new PeopleActivityGenerator(this);

  public Process generateBpelProcess(
          final BpmnMap bpmnMap,
          final TProcess process,
          final QName processQName,
          final List<TFlowNode> sequenceNodes,
          final Map<String, Definition> partnerWSDLs) throws JAXBException, IOException, SAXException {
    final Process p = new Process();
    p.setName(process.getName());
    p.setTargetNamespace(processQName.getNamespaceURI());

    // create bpel main sequence
    final Sequence sequence = new Sequence();

    if (bpmnMap.getLinks().isEmpty()) {
      p.setSequence(sequence);
    } else {
      final Links links = this.generateLinks(bpmnMap);
      final Flow mainFlow = new Flow();
      mainFlow.setLinks(links);
      mainFlow.getAssignsAndCompensatesAndCompensateScopes().add(sequence);
      p.setFlow(mainFlow);
    }

    // fill the sequence
    for (final TFlowNode node : sequenceNodes) {
      final TActivity activity = this.createBpelActivity(node, bpmnMap, processQName, partnerWSDLs);
      sequence.getAssignsAndCompensatesAndCompensateScopes().add(activity);
    }

    if (!bpmnMap.getParticipants().isEmpty()) {
      // create partner links for process participants
      final PartnerLinks partnerLinks = this.generatePartnerLinks(bpmnMap, processQName);
      p.setPartnerLinks(partnerLinks);

    }

    if (!bpmnMap.getDataObjects().isEmpty()) {
      // create variables
      final Variables variables = this.generateProcessVariables(bpmnMap, processQName);
      p.setVariables(variables);
    }

    // generate extensions parts
    this.generateExtensionsDeclaration(p);

    // Generate WSDL for onEvent
    WSDLGenerator.generateOnEventWsdlDefinition(bpmnMap, processQName, onEventGenerator.getOperations(), partnerWSDLs);

    // Add imports for wsdls
    for (final Entry<String, Definition> partnerWSDL : partnerWSDLs.entrySet()) {
      final Import bpelPartnerImport = new Import();
      bpelPartnerImport.setImportType(WSDLGenerator.XMLNS_WSDL);
      bpelPartnerImport.setNamespace(partnerWSDL.getValue().getTargetNamespace());
      bpelPartnerImport.setLocation(partnerWSDL.getKey());

      p.getImports().add(bpelPartnerImport);
    }
    return p;
  }

  private Variables generateProcessVariables(final BpmnMap bpmnMap, final QName processQName) {
    final Variables variables = new Variables();
    for (final TDataObject dataObject : bpmnMap.getDataObjects().values()) {
      final Variable variable = new Variable();
      variable.setName(dataObject.getName());

      final QName itemDefinitionQName = dataObject.getItemSubjectRef();
      final TItemDefinition itemDefinition = bpmnMap.getItemDefinitions().get(itemDefinitionQName);

      if (itemDefinition == null) {
        throw new TransformerException("Missing itemDefinition: " + itemDefinitionQName);
      }
      if (itemDefinition.getStructureRef() != null) {
        variable.setType(itemDefinition.getStructureRef());
      } else {
        final List<Element> outItemElements = itemDefinition.getExtensionElements().getAnies();
        for (final Element elt : outItemElements) {
          if (elt.getNamespaceURI().equals(XMLConstants.W3C_XML_SCHEMA_NS_URI)) {
            if (elt.getLocalName().equals("element")) {
              variable.setElement(new QName(processQName.getNamespaceURI(), elt.getAttribute("name")));
            } else {
              variable.setType(new QName(processQName.getNamespaceURI(), elt.getAttribute("name")));
            }
          }
        }
      }

      variables.getVariables().add(variable);
    }
    return variables;
  }

  private PartnerLinks generatePartnerLinks(final BpmnMap bpmnMap, final QName processQName) {
    final PartnerLinks partnerLinks = new PartnerLinks();

    for (final Entry<QName, Participant> entry : bpmnMap.getParticipants().entrySet()) {
      final QName participantQName = entry.getKey();
      final Participant participant = entry.getValue();

      // add wsdl informations
      for (final QName interfaceQName : participant.getInterfaceReves()) {
        // create partner link
        final PartnerLink partnerLink = new PartnerLink();
        // set partnerlink name
        partnerLink.setName(bpmnMap.getPartnerLinkName(participantQName, interfaceQName));

        // set partnerlinktype (same as in wsdl)
        partnerLink.setPartnerLinkType(
                new QName(
                        processQName.getNamespaceURI(),
                        bpmnMap.getPartnerLinkTypeName(participantQName)));

        // set role
        if (processQName.equals(participant.getProcessRef())) {
          partnerLink.setMyRole(bpmnMap.getRoleName(interfaceQName));
        } else {
          partnerLink.setPartnerRole(bpmnMap.getRoleName(interfaceQName));
        }
        // add partner link
        partnerLinks.getPartnerLinks().add(partnerLink);
      }
    }
    return partnerLinks;
  }

  private Links generateLinks(final BpmnMap bpmnMap) {
    final Links links = new Links();

    for (final LinkComponent linkComponent : bpmnMap.getLinks()) {
      final Link link = new Link();
      link.setName(linkComponent.getName());
      links.getLinks().add(link);
    }
    return links;
  }

  /**
   * Create bpel activity for the flowNode.
   */
  TActivity createBpelActivity(final TFlowNode node, final BpmnMap bpmnMap, final QName processQName, Map<String, Definition> partnerWsdls) throws JAXBException, IOException, SAXException {
    // create activity
    final TActivity activity = this.createBpelActivityWithNoLinks(node, bpmnMap, processQName, partnerWsdls);
    // add targets
    final List<LinkComponent> incomingLinks = bpmnMap.getIncomingLinks(node);
    if (!incomingLinks.isEmpty()) {
      final Targets targets = new Targets();
      activity.setTargets(targets);
      for (final LinkComponent linkComponent : incomingLinks) {
        final Target target = new Target();
        target.setLinkName(linkComponent.getName());
        targets.getTargets().add(target);
      }
      if (incomingLinks.size() > 1) {
        // if more than one incoming link, generate joinCondition
        final JoinConditionComponent joinConditionComponent = bpmnMap.getJoinCondition(node);
        final String joinConditionExpression = joinConditionComponent.getExpression();
        final TCondition condition = new TCondition();
        condition.getContent().add(joinConditionExpression);
        targets.setJoinCondition(condition);
      }
    }
    // add sources
    final List<LinkComponent> outgoingLinks = bpmnMap.getOutgoingLinks(node);
    if (!outgoingLinks.isEmpty()) {
      final Sources sources = new Sources();
      activity.setSources(sources);
      for (final LinkComponent linkComponent : outgoingLinks) {
        final Source source = new Source();
        source.setLinkName(linkComponent.getName());
        if (linkComponent.getTransitionCondition() != null) {
          final TCondition condition = new TCondition();
          condition.getContent().addAll(this.getXPathExpression(linkComponent.getTransitionCondition().getContent()));
          source.setTransitionCondition(condition);
        }
        sources.getSources().add(source);
      }
    }
    return activity;
  }

  private TActivity createBpelActivityWithNoLinks(final TFlowNode node, final BpmnMap bpmnMap, final QName processQName, Map<String, Definition> partnerWsdls) throws JAXBException, IOException, SAXException {
    if (node instanceof SequenceComponent) {
      final SequenceComponent sequenceComponent = (SequenceComponent) node;
      return this.sequenceGenerator.createSequenceActivity(bpmnMap, sequenceComponent, processQName, partnerWsdls);
    }
    if (node instanceof FlowComponent) {
      final FlowComponent flowComponent = (FlowComponent) node;
      return this.flowGenerator.createFlowActivity(bpmnMap, flowComponent, processQName, partnerWsdls);
    }
    if (node instanceof PickComponent) {
      final PickComponent pickComponent = (PickComponent) node;
      return this.pickGenerator.createPickActivity(bpmnMap, pickComponent, processQName, partnerWsdls);
    }
    if (node instanceof SwitchComponent) {
      final SwitchComponent switchComponent = (SwitchComponent) node;
      return this.ifGenerator.createSwitchActivity(bpmnMap, switchComponent, processQName, partnerWsdls);
    }
    if (node instanceof WhileComponent) {
      final WhileComponent whileComponent = (WhileComponent) node;
      return this.whileGenerator.createWhileActivity(bpmnMap, whileComponent, processQName, partnerWsdls);
    }
    if (node instanceof RepeatComponent) {
      final RepeatComponent repeatComponent = (RepeatComponent) node;
      return this.whileGenerator.createRepeatActivity(bpmnMap, repeatComponent, processQName, partnerWsdls);
    }
    if (node instanceof RepeatWhileComponent) {
      final RepeatWhileComponent repeatComponent = (RepeatWhileComponent) node;
      return this.whileGenerator.createRepeatWhileActivity(bpmnMap, repeatComponent, processQName, partnerWsdls);
    }
    if (node instanceof OnEventComponent) {
      final OnEventComponent onEventComponent = (OnEventComponent) node;
      return this.onEventGenerator.createOnEventActivity(bpmnMap, onEventComponent, processQName, partnerWsdls);
    }
    if (node instanceof EmptyComponent) {
      final Empty empty = new Empty();
      final Documentation doc = new Documentation();
      doc.getContent().add(node.getId() + " (generated by transformation)");
      empty.getDocumentations().add(doc);
      return empty;
    }
    if (node instanceof TServiceTask) {
      final TServiceTask serviceTask = (TServiceTask) node;
      return this.invokeGenerator.createInvokeActivity(bpmnMap, serviceTask, processQName);
    }
    if (node instanceof TUserTask) {
      final TUserTask userTask = (TUserTask) node;
      return this.peopleActivityGenerator.createPeopleActivity(bpmnMap, userTask, processQName, partnerWsdls);
    }
    if (node instanceof TCatchEvent) {
      final TCatchEvent event = (TCatchEvent) node;

      for (final JAXBElement<? extends TEventDefinition> eventDefinition : event.getEventDefinitions()) {
        if (eventDefinition.getValue() instanceof TMessageEventDefinition) {
          final TMessageEventDefinition messageEventDefinition = (TMessageEventDefinition) eventDefinition.getValue();
          return this.receiveGenerator.createReceiveActivity(bpmnMap, processQName, event, messageEventDefinition);
        } else if (eventDefinition.getValue() instanceof TTimerEventDefinition) {
          final TTimerEventDefinition timerEventDefinition = (TTimerEventDefinition) eventDefinition.getValue();
          return this.waitGenerator.createWaitActivity(event, timerEventDefinition);
        }
      }
    }
    if (node instanceof TThrowEvent) {
      final TThrowEvent event = (TThrowEvent) node;

      for (final JAXBElement<? extends TEventDefinition> eventDefinition : event.getEventDefinitions()) {
        if (eventDefinition.getValue() instanceof TMessageEventDefinition) {
          final TMessageEventDefinition messageEventDefinition = (TMessageEventDefinition) eventDefinition.getValue();
          return this.replyGenerator.createReplyActivity(bpmnMap, processQName, event, messageEventDefinition);
        }
      }
    }
    final Empty empty = new Empty();
    empty.setName(node.getId());
    final Documentation doc = new Documentation();
    doc.getContent().add(node.getName());
    empty.getDocumentations().add(doc);
    return empty;
  }


  Object getOutputData(
          final List<DataOutputAssociation> dataOutputAssociations,
          final OperationReference operationReference,
          final Assign assign,
          final boolean inOperation,
          final BpmnMap bpmnMap) {
    boolean hasAssignments = false;
    for (final DataOutputAssociation dataOutputAssociation : dataOutputAssociations) {
      if (!dataOutputAssociation.getAssignments().isEmpty()) {
        hasAssignments = true;
      }
    }
    if (!hasAssignments) {
      final FromParts fromParts = new FromParts();
      for (final DataOutputAssociation dataOutputAssociation : dataOutputAssociations) {
        if (dataOutputAssociation.getSourceReves().size() != 1) {
          throw new TransformerException("data output association with multiple sources without assignement not supported");
        }
        final FromPart fromPart = new FromPart();
        fromPart.setToVariable(this.getItemAwareElementName(dataOutputAssociation.getTargetRef()));
        fromPart.setPart(this.getItemAwareElementName(dataOutputAssociation.getSourceReves().get(0).getValue()));
        fromParts.getFromParts().add(fromPart);
      }
      return fromParts;
    } else {
      final Variable outVariable = new Variable();
      outVariable.setName(Misc.getHumanReadableId("outVar"));

      if (inOperation) {
        if (operationReference.getPortType() == null) {
          final QName messageQName = operationReference.getOperation().getInMessageRef();
          final TMessage msg = bpmnMap.getMessages().get(messageQName);
          if (msg == null) {
            throw new TransformerException("message " + messageQName + " not found !");
          }
          outVariable.setMessageType(new QName(messageQName.getNamespaceURI(), msg.getName(), messageQName.getPrefix()));
        } else {
          final QName messageQName =
                  operationReference
                          .getPortType()
                          .getOperation(operationReference.getOperation().getName(), null, null)
                          .getInput()
                          .getMessage()
                          .getQName();
          outVariable.setMessageType(messageQName);
        }
      } else {
        if (operationReference.getPortType() == null) {
          final QName messageQName = operationReference.getOperation().getOutMessageRef();
          final TMessage msg = bpmnMap.getMessages().get(messageQName);
          if (msg == null) {
            throw new TransformerException("message " + messageQName + " not found !");
          }
          outVariable.setMessageType(new QName(messageQName.getNamespaceURI(), msg.getName(), messageQName.getPrefix()));
        } else {
          final QName messageQName =
                  operationReference
                          .getPortType()
                          .getOperation(operationReference.getOperation().getName(), null, null)
                          .getOutput()
                          .getMessage()
                          .getQName();
          outVariable.setMessageType(messageQName);
        }
      }

      for (final DataOutputAssociation dataOutputAssociation : dataOutputAssociations) {
        if (dataOutputAssociation.getAssignments().isEmpty()) {
          if (dataOutputAssociation.getSourceReves().size() != 1) {
            throw new TransformerException("data output association with multiple sources without assignement not supported");
          }
          final Copy copy = new Copy();
          final To to = new To();
          to.setVariable(this.getItemAwareElementName(dataOutputAssociation.getTargetRef()));

          final From from = new From();
          from.setVariable(outVariable.getName());
          from.setPart(this.getItemAwareElementName(dataOutputAssociation.getSourceReves().get(0).getValue()));

          copy.setFrom(from);
          copy.setTo(to);
          assign.getCopiesAndExtensionAssignOperations().add(copy);
        } else {
          // with assignments
          for (final Assignment assignment : dataOutputAssociation.getAssignments()) {
            final Copy copy = new Copy();
            final From from = new From();
            from.getContent().addAll(
                    this.getXPathExpression(assignment.getFrom().getContent(), null, outVariable.getName())
            );

            final To to = new To();
            to.getContent().addAll(
                    this.getXPathExpression(assignment.getTo().getContent(), null, outVariable.getName()));

            copy.setFrom(from);
            copy.setTo(to);
            assign.getCopiesAndExtensionAssignOperations().add(copy);
          }
        }
      }
      return outVariable;
    }
  }

  OperationReference getProcessOperationReference(
          final BpmnMap bpmnMap,
          final QName processQName,
          final QName operationQName) {
    for (final Entry<QName, Participant> entry : bpmnMap.getParticipants().entrySet()) {
      final Participant participant = entry.getValue();
      if (processQName.equals(participant.getProcessRef())) {
        final QName participantQName = entry.getKey();
        return this.getOperationReference(bpmnMap, operationQName, participantQName);
      }
    }
    return null;
  }


  OperationReference getOperationReference(
          final BpmnMap bpmnMap,
          final QName operationQName,
          final QName participantQName) {
    for (final QName interfaceQName : bpmnMap.getParticipants().get(participantQName).getInterfaceReves()) {
      if (interfaceQName.getNamespaceURI().equals(operationQName.getNamespaceURI())) {
        final TInterface tInterface = bpmnMap.getInterfaces().get(interfaceQName);
        if (tInterface == null) {
          throw new TransformerException("Missing interface: " + interfaceQName);
        }
        for (final Operation operation : tInterface.getOperations()) {
          if (operation.getId().equals(operationQName.getLocalPart())) {
            if (tInterface.getImplementationRef() != null) {
              return new OperationReference(
                      bpmnMap.getPortType(tInterface.getImplementationRef()),
                      bpmnMap.getPartnerLinkName(participantQName, interfaceQName),
                      operation);
            } else {
              return new OperationReference(
                      new QName(interfaceQName.getNamespaceURI(), tInterface.getName(), interfaceQName.getPrefix()),
                      bpmnMap.getPartnerLinkName(participantQName, interfaceQName),
                      operation);
            }
          }
        }
      }
    }
    return null;
  }

  String getItemAwareElementName(final Object itemAwareElement) {
    if (itemAwareElement instanceof DataInput) {
      return ((DataInput) itemAwareElement).getName();
    } else if (itemAwareElement instanceof DataOutput) {
      return ((DataOutput) itemAwareElement).getName();
    } else if (itemAwareElement instanceof TDataObject) {
      return ((TDataObject) itemAwareElement).getName();
    } else if (itemAwareElement instanceof Property) {
      return ((Property) itemAwareElement).getName();
    }
    throw new TransformerException("Unsupported item type: " + itemAwareElement.getClass().getName());
  }

  Object getInputData(
          final List<DataInputAssociation> dataInputAssociations,
          final OperationReference invokeOperationReference,
          final Assign inAssign,
          final boolean inOperation,
          final BpmnMap bpmnMap) {
    boolean hasAssignments = false;
    for (final DataInputAssociation dataInputAssociation : dataInputAssociations) {
      if (!dataInputAssociation.getAssignments().isEmpty()) {
        hasAssignments = true;
      }
    }
    if (!hasAssignments) {
      final ToParts toParts = new ToParts();
      for (final DataInputAssociation dataInputAssociation : dataInputAssociations) {
        if (dataInputAssociation.getSourceReves().size() != 1) {
          throw new TransformerException("data input association with multiple sources without assignement not supported");
        }
        final ToPart toPart = new ToPart();
        toPart.setFromVariable(this.getItemAwareElementName(dataInputAssociation.getSourceReves().get(0).getValue()));
        toPart.setPart(this.getItemAwareElementName(dataInputAssociation.getTargetRef()));
        toParts.getToParts().add(toPart);
      }
      return toParts;
    } else {
      final Variable inVariable = new Variable();
      inVariable.setName(Misc.getHumanReadableId("inVar"));

      if (inOperation) {
        if (invokeOperationReference.getPortType() == null) {
          final QName messageQName = invokeOperationReference.getOperation().getInMessageRef();
          final TMessage msg = bpmnMap.getMessages().get(messageQName);
          if (msg == null) {
            throw new TransformerException("message " + messageQName + " not found !");
          }
          inVariable.setMessageType(new QName(messageQName.getNamespaceURI(), msg.getName(), messageQName.getPrefix()));
        } else {
          final QName messageQName =
                  invokeOperationReference
                          .getPortType()
                          .getOperation(invokeOperationReference.getOperation().getName(), null, null)
                          .getInput()
                          .getMessage()
                          .getQName();
          inVariable.setMessageType(messageQName);
        }
      } else {
        if (invokeOperationReference.getPortType() == null) {
          final QName messageQName = invokeOperationReference.getOperation().getOutMessageRef();
          final TMessage msg = bpmnMap.getMessages().get(messageQName);
          if (msg == null) {
            throw new TransformerException("message " + messageQName + " not found !");
          }
          inVariable.setMessageType(new QName(messageQName.getNamespaceURI(), msg.getName(), messageQName.getPrefix()));
        } else {
          final QName messageQName =
                  invokeOperationReference
                          .getPortType()
                          .getOperation(invokeOperationReference.getOperation().getName(), null, null)
                          .getOutput()
                          .getMessage()
                          .getQName();
          inVariable.setMessageType(messageQName);
        }
      }
      for (final DataInputAssociation dataInputAssociation : dataInputAssociations) {
        if (dataInputAssociation.getAssignments().isEmpty()) {
          if (dataInputAssociation.getSourceReves().size() != 1) {
            throw new TransformerException("data input association with multiple sources without assignement not supported");
          }
          final Copy copy = new Copy();
          final From from = new From();
          from.setVariable(this.getItemAwareElementName(dataInputAssociation.getSourceReves().get(0).getValue()));

          final To to = new To();
          to.setVariable(inVariable.getName());
          to.setPart(this.getItemAwareElementName(dataInputAssociation.getTargetRef()));

          copy.setFrom(from);
          copy.setTo(to);
          inAssign.getCopiesAndExtensionAssignOperations().add(copy);
        } else {
          // with assignments
          for (final Assignment assignment : dataInputAssociation.getAssignments()) {
            final Copy copy = new Copy();
            final From from = new From();
            from.getContent().addAll(
                    this.getXPathExpression(assignment.getFrom().getContent(), inVariable.getName(), null)
            );

            final To to = new To();
            to.getContent().addAll(this.getXPathExpression(assignment.getTo().getContent(), inVariable.getName(), null));

            copy.setFrom(from);
            copy.setTo(to);
            inAssign.getCopiesAndExtensionAssignOperations().add(copy);
          }
        }
      }
      return inVariable;
    }
  }

  /**
   * Transform BPMN expressions to BPEL expressions
   *
   * @param content
   * @return
   */
  Collection<? extends Object> getXPathExpression(final List<Object> content) {
    return this.getXPathExpression(content, null, null);
  }


  Collection<? extends Object> getXPathExpression(
      final List<Object> content,
      final String inVarName,
      final String outVarName) {
    final ListIterator<Object> iterator = content.listIterator();
    while (iterator.hasNext()) {
      final Object o = iterator.next();
      if (o instanceof String) {
        String expr = (String) o;
        if (expr.contains("getDataInput") && inVarName == null) {
          throw new TransformerException("No input variable to replace getDataInput in " + expr);
        }
        if (expr.contains("getDataOutput") && outVarName == null) {
          throw new TransformerException("No input variable to replace getDataOutput in " + expr);
        }
        iterator.remove();
        // getDataObject("varName") -> $varName
        expr = expr.replaceAll("getDataObject\\(\"(\\w*)\"\\)", "\\$$1");
        expr = expr.replaceAll("getDataObject\\('(\\w*)'\\)", "\\$$1");
        // getDataInput("varName") -> $inVar.varName
        expr = expr.replaceAll("getDataInput\\(\"(\\w*)\"\\)", "\\$" + inVarName + ".$1");
        expr = expr.replaceAll("getDataInput\\('(\\w*)'\\)", "\\$" + inVarName + ".$1");
        // getDataOutput("varName") -> $outVar.varName
        expr = expr.replaceAll("getDataOutput\\(\"(\\w*)\"\\)", "\\$" + outVarName + ".$1");
        expr = expr.replaceAll("getDataOutput\\('(\\w*)'\\)", "\\$" + outVarName + ".$1");
        iterator.add(expr);
      }
    }
    return content;
  }

  private void generateExtensionsDeclaration(Process process) {
    if (this.peopleActivityGenerator.isAddExtensionDeclaration()) {
      Extensions extensions = new Extensions();
      process.setExtensions(extensions);

      Extension wshtExtension = new Extension();
      wshtExtension.setMustUnderstand(TBoolean.YES);
      wshtExtension.setNamespace("http://www.example.org/WS-HT");
      extensions.getExtensions().add(wshtExtension);

      Extension b4pExtension = new Extension();
      b4pExtension.setMustUnderstand(TBoolean.YES);
      b4pExtension.setNamespace("http://www.example.org/BPEL4People");
      extensions.getExtensions().add(b4pExtension);

    }
  }
}
