/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel.generators;

import org.ow2.orchestra.jaxb.bpel.Documentation;
import org.ow2.orchestra.jaxb.bpel.TDurationExpr;
import org.ow2.orchestra.jaxb.bpel.Until;
import org.ow2.orchestra.jaxb.bpel.Wait;
import org.ow2.orchestra.jaxb.bpmn.TCatchEvent;
import org.ow2.orchestra.jaxb.bpmn.TTimerEventDefinition;


/**
 * @author Guillaume Porcher
 */
class WaitGenerator {
  private final BpelGenerator bpelGenerator;

  WaitGenerator() {
    this.bpelGenerator = new BpelGenerator();
  }

  /**
   * @param bpelGenerator
   */
  WaitGenerator(final BpelGenerator bpelGenerator) {
    this.bpelGenerator = bpelGenerator;
  }

  Wait createWaitActivity(final TCatchEvent event, final TTimerEventDefinition timerDef) {
    // create wait
    final Wait wait = new Wait();
    wait.setName(event.getId());
    final Documentation doc = new Documentation();
    doc.getContent().add(event.getName());
    wait.getDocumentations().add(doc);
    if (timerDef.getTimeDuration() != null) {
      final TDurationExpr duration = new TDurationExpr();
      duration.getContent().addAll(this.bpelGenerator.getXPathExpression(timerDef.getTimeDuration().getContent()));
      wait.setFor(duration);
    } else if (timerDef.getTimeDate() != null) {
      final Until until = new Until();
      until.getContent().addAll(this.bpelGenerator.getXPathExpression(timerDef.getTimeDate().getContent()));
      wait.setUntil(until);
    }
    return wait;
  }

}
